using UnityEngine;
using System.Collections;
using System.Text.RegularExpressions;

public class BlockPosition
{
    public int x;
    public int y;

    public BlockPosition(int setX, int setY)
    {
        x = setX;
        y = setY;
    }

    public override string ToString()
    {
        return "(" + x + ", " + y + ")";
    }
}


public class BlockManager : MonoBehaviour
{
    public GameObject BlockPrefab;
    public int LevelWidth = 20;
    public int LevelHeight = 15;
    public TextAsset WordFile;
    public Transform CurrentWordHolder;
    public GUIStyle ScoreDisplayStyle;
    public GUIStyle GameOverStyle;
    public GUIStyle GameOverButtonStyle;
    public AudioClip MadeWordSound;
    public AudioClip SelectLetterSound;
    public AudioClip BadWordSound;

    private GameObject[,] stationaryBlocks;

    private ArrayList selectionBlocks = new ArrayList();
    private ArrayList selectionBlockObjects = new ArrayList();
    private BlockPosition lastSelection;

    private Piece fallingPiece = null;
    private ArrayList fallingPieceGameObjects;
    private BlockPosition fallingPiecePosition;

    private float timeUntilNextDrop;

    private ArrayList wordList;

    private int score;
    private bool gameOver = false;

    public float TimeBetweenLevels = 30.0f;
    public int currentLevel = 0;
    private float timeToNextLevel = 0.0f;

    public float TimeBetweenHorizontalShift = 0.25f;
    private float timeSinceHorizontalShift = 0.0f;


    private static string RemoveSpecialCharacters(string str)
    {
        return Regex.Replace(str, "[^a-z]+", "");
    }

    void Start()
    {
        score = 0;
        string wordFileString = WordFile.text;
        string[] wordFileSplit = wordFileString.Split('\n');
        wordList = new ArrayList();
        foreach (string s in wordFileSplit)
        {
            wordList.Add(RemoveSpecialCharacters(s));
        }

        timeUntilNextDrop = 1.0f;
        fallingPieceGameObjects = new ArrayList();
        stationaryBlocks = new GameObject[LevelWidth, LevelHeight];
        for (int x = 0; x < LevelWidth; ++x)
        {
            for (int y = 0; y < LevelHeight; ++y)
            {
                stationaryBlocks[x, y] = null;
            }
        }

        lastSelection = new BlockPosition(-1, -1);

        AddNewFallingPiece();
    }

    private Vector3 Get3DPosition(BlockPosition position)
    {
        return new Vector3(position.x, position.y, 0.0f);
    }

    private GameObject GetBlockGameObjectAtPosition(BlockPosition position)
    {
        return stationaryBlocks[position.x, position.y];
    }

    private GameObject SetBlock(BlockPosition position)
    {
        GameObject newBlock = (GameObject)Instantiate(BlockPrefab, Get3DPosition(position), Quaternion.LookRotation(new Vector3(1.0f, 0.0f, 0.0f), Vector3.up));
        GameObject oldGameObject = GetBlockGameObjectAtPosition(position);
        if (oldGameObject != null)
        {
            DestroyImmediate(oldGameObject);
        }

        stationaryBlocks[position.x, position.y] = newBlock;
        return newBlock;
    }

    private void RemoveBlock(BlockPosition position)
    {
        if (stationaryBlocks[position.x, position.y] != null)
        {
            DestroyImmediate(stationaryBlocks[position.x, position.y]);
        }
    }

    private bool HasBlock(BlockPosition position)
    {
        if (stationaryBlocks[position.x, position.y] == null)
        {
            return false;
        }

        return true;
    }

    private void UpdateBlockPosition(BlockPosition position)
    {
        GameObject objToUpdate = GetBlockGameObjectAtPosition(position);
        if (objToUpdate != null)
        {
            objToUpdate.transform.position = Get3DPosition(position);
        }
    }

    private bool BlockPositionInLevel(BlockPosition position)
    {
        if (position.x < 0 || position.x >= LevelWidth)
        {
            return false;
        }
        else if (position.y < 0 || position.y >= LevelHeight)
        {
            return false;
        }

        return true;
    }

    private void SwapBlocks(BlockPosition positionA, BlockPosition positionB)
    {
        GameObject tempObjectA = GetBlockGameObjectAtPosition(positionA);
        stationaryBlocks[positionA.x, positionA.y] = GetBlockGameObjectAtPosition(positionB);
        stationaryBlocks[positionB.x, positionB.y] = tempObjectA;
        UpdateBlockPosition(positionA);
        UpdateBlockPosition(positionB);
    }

    private void SwapPiece(BlockPosition oldPosition1, BlockPosition newPosition1, BlockPosition oldPosition2, BlockPosition newPosition2, BlockPosition oldPosition3, BlockPosition newPosition3, BlockPosition oldPosition4, BlockPosition newPosition4)
    {
        GameObject tempObject1 = GetBlockGameObjectAtPosition(oldPosition1);
        GameObject tempObject2 = GetBlockGameObjectAtPosition(oldPosition2);
        GameObject tempObject3 = GetBlockGameObjectAtPosition(oldPosition3);
        GameObject tempObject4 = GetBlockGameObjectAtPosition(oldPosition4);

        stationaryBlocks[oldPosition1.x, oldPosition1.y] = null;
        stationaryBlocks[oldPosition2.x, oldPosition2.y] = null;
        stationaryBlocks[oldPosition3.x, oldPosition3.y] = null;
        stationaryBlocks[oldPosition4.x, oldPosition4.y] = null;

        stationaryBlocks[newPosition1.x, newPosition1.y] = tempObject1;
        stationaryBlocks[newPosition2.x, newPosition2.y] = tempObject2;
        stationaryBlocks[newPosition3.x, newPosition3.y] = tempObject3;
        stationaryBlocks[newPosition4.x, newPosition4.y] = tempObject4;

        UpdateBlockPosition(oldPosition1);
        UpdateBlockPosition(oldPosition2);
        UpdateBlockPosition(oldPosition3);
        UpdateBlockPosition(oldPosition4);
        UpdateBlockPosition(newPosition1);
        UpdateBlockPosition(newPosition2);
        UpdateBlockPosition(newPosition3);
        UpdateBlockPosition(newPosition4);
    }

    private BlockPosition GetPositionFromGameObject(GameObject blockObject)
    {
        for (int x = 0; x < LevelWidth; ++x)
        {
            for (int y = 0; y < LevelHeight; ++y)
            {
                if (stationaryBlocks[x, y] == blockObject)
                {
                    return new BlockPosition(x, y);
                }
            }
        }

        return new BlockPosition(-1, -1);
    }

    private void BottomOutAllBlocks()
    {
        for (int x = 0; x < LevelWidth; ++x)
        {
            for (int y = 1; y < LevelHeight; ++y)
            {
                if (fallingPieceGameObjects.Contains(stationaryBlocks[x, y]))
                {
                    continue;
                }

                if (stationaryBlocks[x, y] != null && stationaryBlocks[x, y - 1] == null)
                {
                    int newYPosition = y - 1;
                    while (newYPosition >= 0 && stationaryBlocks[x, newYPosition] == null)
                    {
                        newYPosition--;
                    }
                    newYPosition++;
                    SwapBlocks(new BlockPosition(x, y), new BlockPosition(x, newYPosition));
                }
            }
        }
    }

    private string GetWordFromSelection()
    {
        string builtString = "";
        foreach (GameObject selectionBlock in selectionBlocks)
        {
            Block blockInfo = (Block)selectionBlock.GetComponent("Block");
            if (blockInfo != null)
            {
                builtString += blockInfo.Letter;
            }
        }
        return builtString;
    }

    private BlockPosition GetFallingPieceStartPosition()
    {
        return new BlockPosition((int)(LevelWidth / 2) - 3, LevelHeight - 5);
    }

    private void AddFallingPiecesToStationary()
    {
        fallingPiece = null;
        fallingPieceGameObjects.Clear();
    }

    private bool FallingPieceFits(BlockPosition testPosition, bool tryRotate)
    {
        for (int x = 0; x < 5; ++x)
        {
            for (int y = 0; y < 5; ++y)
            {
                bool fallingPieceHasBlock = false;
                if (tryRotate == false && fallingPiece.HasBlockAt(x, y))
                {
                    fallingPieceHasBlock = true;
                }
                else if (tryRotate == true && fallingPiece.NextRotationHasBlockAt(x, y))
                {
                    fallingPieceHasBlock = true;
                }

                if (fallingPieceHasBlock == true)
                {
                    if (BlockPositionInLevel(new BlockPosition(testPosition.x + x, testPosition.y + y)) == false)
                    {
                        return false;
                    }

                    GameObject blockingGameObject = GetBlockGameObjectAtPosition(new BlockPosition(testPosition.x + x, testPosition.y + y));
                    if (blockingGameObject != null && fallingPieceGameObjects.Contains(blockingGameObject) == false)
                    {
                        return false;
                    }
                }
            }
        }

        return true;
    }

    private void SetFallingPieceBlocks()
    {
        for (int x = 0; x < 5; ++x)
        {
            for (int y = 0; y < 5; ++y)
            {
                if (fallingPiece.HasBlockAt(x, y) == true)
                {
                    GameObject newBlock = SetBlock(new BlockPosition(fallingPiecePosition.x + x, fallingPiecePosition.y + y));
                    fallingPieceGameObjects.Add(newBlock);
                }
            }
        }
    }

    private void AddNewFallingPiece()
    {
        if (fallingPiece != null)
        {
            return;
        }

        fallingPiece = new Piece();
        fallingPiecePosition = GetFallingPieceStartPosition();

        if (FallingPieceFits(fallingPiecePosition, false) == false)
        {
            gameOver = true;
            FillLevelWithRandomBlocks();
            return;
        }

        SetFallingPieceBlocks();
    }

    private void FillLevelWithRandomBlocks()
    {
        for (int x = 0; x < LevelWidth; ++x)
        {
            for (int y = 0; y < LevelHeight; ++y)
            {
                if (stationaryBlocks[x, y] == null)
                {
                    SetBlock(new BlockPosition(x, y));
                }
            }
        }
    }

    private void DropFallingPiece()
    {
        if (FallingPieceFits(new BlockPosition(fallingPiecePosition.x, fallingPiecePosition.y - 1), false))
        {
            for (int y = 0; y < 5; ++y)
            {
                for (int x = 0; x < 5; ++x)
                {
                    if (fallingPiece.HasBlockAt(x, y) == true)
                    {
                        BlockPosition originalPosition = new BlockPosition(fallingPiecePosition.x + x, fallingPiecePosition.y + y);
                        BlockPosition newPosition = new BlockPosition(fallingPiecePosition.x + x, fallingPiecePosition.y + y - 1);

                        if (BlockPositionInLevel(originalPosition) && BlockPositionInLevel(newPosition))
                        {
                            SwapBlocks(originalPosition, newPosition);
                        }
                    }
                }
            }
            fallingPiecePosition.y -= 1;
        }
        else
        {
            AddFallingPiecesToStationary();
            BottomOutAllBlocks();
            AddNewFallingPiece();
        }
    }

    private void RotateFallingPiece()
    {
        if (FallingPieceFits(fallingPiecePosition, true) == true)
        {
            BlockPosition oldPosition1 = fallingPiece.GetPositionOfIndex(1, false);
            BlockPosition oldPosition2 = fallingPiece.GetPositionOfIndex(2, false);
            BlockPosition oldPosition3 = fallingPiece.GetPositionOfIndex(3, false);
            BlockPosition oldPosition4 = fallingPiece.GetPositionOfIndex(4, false);

            BlockPosition newPosition1 = fallingPiece.GetPositionOfIndex(1, true);
            BlockPosition newPosition2 = fallingPiece.GetPositionOfIndex(2, true);
            BlockPosition newPosition3 = fallingPiece.GetPositionOfIndex(3, true);
            BlockPosition newPosition4 = fallingPiece.GetPositionOfIndex(4, true);

            BlockPosition totalOldPosition1 = new BlockPosition(fallingPiecePosition.x + oldPosition1.x, fallingPiecePosition.y + oldPosition1.y);
            BlockPosition totalOldPosition2 = new BlockPosition(fallingPiecePosition.x + oldPosition2.x, fallingPiecePosition.y + oldPosition2.y);
            BlockPosition totalOldPosition3 = new BlockPosition(fallingPiecePosition.x + oldPosition3.x, fallingPiecePosition.y + oldPosition3.y);
            BlockPosition totalOldPosition4 = new BlockPosition(fallingPiecePosition.x + oldPosition4.x, fallingPiecePosition.y + oldPosition4.y);

            BlockPosition totalNewPosition1 = new BlockPosition(fallingPiecePosition.x + newPosition1.x, fallingPiecePosition.y + newPosition1.y);
            BlockPosition totalNewPosition2 = new BlockPosition(fallingPiecePosition.x + newPosition2.x, fallingPiecePosition.y + newPosition2.y);
            BlockPosition totalNewPosition3 = new BlockPosition(fallingPiecePosition.x + newPosition3.x, fallingPiecePosition.y + newPosition3.y);
            BlockPosition totalNewPosition4 = new BlockPosition(fallingPiecePosition.x + newPosition4.x, fallingPiecePosition.y + newPosition4.y);

            SwapPiece(totalOldPosition1, totalNewPosition1, totalOldPosition2, totalNewPosition2, totalOldPosition3, totalNewPosition3, totalOldPosition4, totalNewPosition4);

            fallingPiece.IncrementRotation();
        }
    }

    private void MoveFallingPieceHorizontal(bool moveLeft)
    {
        if (moveLeft)
        {
            if (FallingPieceFits(new BlockPosition(fallingPiecePosition.x - 1, fallingPiecePosition.y), false) == true)
            {
                for (int x = 0; x < 5; ++x)
                {
                    for (int y = 0; y < 5; ++y)
                    {
                        if (fallingPiece.HasBlockAt(x, y) == true)
                        {
                            BlockPosition originalPosition = new BlockPosition(fallingPiecePosition.x + x, fallingPiecePosition.y + y);
                            BlockPosition newPosition = new BlockPosition(fallingPiecePosition.x + x - 1, fallingPiecePosition.y + y);

                            if (BlockPositionInLevel(originalPosition) && BlockPositionInLevel(newPosition))
                            {
                                SwapBlocks(originalPosition, newPosition);
                            }
                        }
                    }
                }
                fallingPiecePosition.x -= 1;
            }
        }
        else
        {
            if (FallingPieceFits(new BlockPosition(fallingPiecePosition.x + 1, fallingPiecePosition.y), false) == true)
            {
                for (int x = 4; x >= 0; --x)
                {
                    for (int y = 0; y < 5; ++y)
                    {
                        if (fallingPiece.HasBlockAt(x, y) == true)
                        {
                            BlockPosition originalPosition = new BlockPosition(fallingPiecePosition.x + x, fallingPiecePosition.y + y);
                            BlockPosition newPosition = new BlockPosition(fallingPiecePosition.x + x + 1, fallingPiecePosition.y + y);

                            if (BlockPositionInLevel(originalPosition) && BlockPositionInLevel(newPosition))
                            {
                                SwapBlocks(originalPosition, newPosition);
                            }
                        }
                    }
                }
                fallingPiecePosition.x += 1;
            }
        }
    }

    private void ScoreWord(string word)
    {
        score += (int)Mathf.Pow(5, word.Length);
    }

    void OnGUI()
    {
        GUI.Label(new Rect(20.0f, 20.0f, 500.0f, 50.0f), "Score: " + score, ScoreDisplayStyle);

        if (gameOver == true)
        {
            GUI.Box(new Rect(Screen.width / 4.0f, Screen.height / 4.0f, Screen.width / 2.0f, Screen.height / 2.0f), "Game Over Dude (or Dudette)!\n\nFinal Score: " + score, GameOverStyle);
            if (GUI.Button(new Rect(Screen.width / 2.25f, Screen.height / 2.0f, Screen.width / 8.0f, Screen.height / 8.0f), "Restart"/*, GameOverButtonStyle*/))
            {
                Application.LoadLevel(0);
            }
        }

        GUI.Label(new Rect(950.0f, 200.0f, 280.0f, 200.0f), "Instructions:\n-------------\n\nWASD - move, flip, drop pieces.\n\nLeft Click and Drag with your mouse to spell words.\n\nThe longer the word, the more points you'll get.");
    }

    void ConsiderAddingToSelection(GameObject gameObj)
    {
        bool passesCheck = false;
        BlockPosition selectionPosition = GetPositionFromGameObject(gameObj);

        if (fallingPieceGameObjects.Contains(gameObj))
        {
            return;
        }

        if (lastSelection.x == -1 && lastSelection.y == -1)
        {
            passesCheck = true;
        }
        else
        {
            int deltaX = Mathf.Abs(lastSelection.x - selectionPosition.x);
            int deltaY = Mathf.Abs(lastSelection.y - selectionPosition.y);

            if ((deltaX == 0 && deltaY == 1) || (deltaX == 1 && deltaY == 0))
            {
                passesCheck = true;
            }
        }

        if (passesCheck == true)
        {
            Block blockComponent = (Block)gameObj.GetComponent("Block");
            if (blockComponent != null)
            {
                blockComponent.Activate();
            }

            selectionBlocks.Add(gameObj);
            audio.PlayOneShot(SelectLetterSound);
            lastSelection = selectionPosition;

            GameObject newBlock =
                (GameObject)
                Instantiate(BlockPrefab,
                            CurrentWordHolder.position + new Vector3(1.1f * selectionBlockObjects.Count, 0.0f, 0.0f),
                            Quaternion.LookRotation(new Vector3(1.0f, 0.0f, 0.0f), Vector3.up));
            Block newBlockComponent = (Block)newBlock.GetComponent("Block");

            newBlockComponent.SetLetter(blockComponent.Letter);
            newBlockComponent.Activate();
            selectionBlockObjects.Add(newBlock);
        }
    }

    void Update()
    {
        if (gameOver == true)
        {
            return;
        }

        timeToNextLevel += Time.deltaTime;
        if (timeToNextLevel > TimeBetweenLevels)
        {
            currentLevel++;
            if (currentLevel > 19)
            {
                currentLevel = 19;
            }
            timeToNextLevel = 0.0f;
        }

        timeSinceHorizontalShift += Time.deltaTime;

        timeUntilNextDrop -= Time.deltaTime;
        if (timeUntilNextDrop < 0.0f)
        {
            DropFallingPiece();
            timeUntilNextDrop = 1.0f - Mathf.Lerp(0.0f, 1.0f, (float)currentLevel / 20.0f);
        }

        if (Input.GetKey(KeyCode.A) && fallingPiece != null && timeSinceHorizontalShift > TimeBetweenHorizontalShift)
        {
            MoveFallingPieceHorizontal(true);
            timeSinceHorizontalShift = 0.0f;
        }
        else if (Input.GetKey(KeyCode.D) && fallingPiece != null && timeSinceHorizontalShift > TimeBetweenHorizontalShift)
        {
            MoveFallingPieceHorizontal(false);
            timeSinceHorizontalShift = 0.0f;
        }
        else if (Input.GetKeyDown(KeyCode.W) && fallingPiece != null)
        {
            RotateFallingPiece();
        }
        else if (Input.GetKey(KeyCode.S) && fallingPiece != null)
        {
            timeUntilNextDrop -= Time.deltaTime * 5.0f;
        }

        if (Input.GetMouseButton(0))
        {
            Ray mouseClickRay = Camera.main.ScreenPointToRay(Input.mousePosition);
            RaycastHit hit;
            if (Physics.Raycast(mouseClickRay, out hit, Mathf.Infinity) && hit.collider.gameObject.CompareTag("StationaryBlock"))
            {
                if (selectionBlocks.Contains(hit.collider.gameObject) == false)
                {
                    ConsiderAddingToSelection(hit.collider.gameObject);
                }
            }
        }
        else
        {
            if (selectionBlocks.Count > 0)
            {
                string word = GetWordFromSelection();
                word = word.ToLower();
                bool inWordList = wordList.Contains(word);
                if (inWordList == true)
                {
                    ScoreWord(word);

                    foreach (GameObject selectedBlock in selectionBlocks)
                    {
                        BlockPosition blockPosition = GetPositionFromGameObject(selectedBlock);
                        RemoveBlock(blockPosition);
                    }

                    selectionBlocks.Clear();

                    foreach (GameObject selectionBlockObject in selectionBlockObjects)
                    {
                        DestroyImmediate(selectionBlockObject);
                    }
                    selectionBlockObjects.Clear();

                    audio.PlayOneShot(MadeWordSound);

                    BottomOutAllBlocks();
                }
                else
                {
                    foreach (GameObject selectionBlock in selectionBlocks)
                    {
                        Block blockComponent = (Block)selectionBlock.GetComponent("Block");
                        if (blockComponent != null)
                        {
                            blockComponent.Deactivate();
                        }
                    }

                    audio.PlayOneShot(BadWordSound);

                    selectionBlocks.Clear();

                    foreach (GameObject selectionBlockObject in selectionBlockObjects)
                    {
                        DestroyImmediate(selectionBlockObject);
                    }
                    selectionBlockObjects.Clear();
                }
            }

            lastSelection = new BlockPosition(-1, -1);
        }
    }
}
